<?php

/* access.php - SynApp2 authentication and database access control
**
   +----------------------------------------------------------------------+
   | SynApp2 Version 1                                                    |
   +----------------------------------------------------------------------+
   | Copyright (c) 2007 - 2009 Richard Howell. All rights reserved.       |
   +----------------------------------------------------------------------+
   | This source file is subject to version 1.01 of the SynApp2 license,  |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.synapp2.org/license/1_01.txt                              |
   | If you did not receive a copy of the SynApp2 license and are unable  |
   | to obtain it through the world-wide-web, please send a note to       |
   | license@synapp2.org so we can mail you a copy immediately.           |
   +----------------------------------------------------------------------+
   | Authors: Richard Howell                                              |
   |                                                                      |
   +----------------------------------------------------------------------+
**
** http://www.synapp2.org
**
** $Id: access.php,v 1.10 2010/10/08 01:17:01 richard Exp $
*/

class access_base
{
    //// begin private data members

    // private:

    var $m_action = null;
    var $m_dbx = null;
    var $m_custom = null;
    var $m_config = null;
    var $m_link = false;
    var $m_connect_as = null;
    var $m_auth_interface = null;

    //// begin public interface

    // public:

    function logout()
    {
        $this->set_token();
    }

    function is_authorized()
    {
        return $this->m_action->is_login() ||
        ($this->m_custom->is_auth_appid_username($this->get_username()) &&
         $this->m_custom->is_auth_qid_username($this->m_action->get_qid(), $this->get_username()) &&
         $this->m_custom->is_auth_pid_username($this->m_action->get_pid(), $this->get_username()) &&
         $this->m_custom->is_auth_rid_username($this->m_action->get_rid(), $this->get_username()));
    }

    function is_authenticated()
    {
        $token = $this->get_token();

        return !empty($token);
    }

    function select_db($db_name)
    {
        $select_result = false;

        if ($this->db_connect())
        {
            $select_result = $this->m_dbx->select_db($db_name);
        }

        $this->select_db_diagnostic($db_name, $select_result);

        return $select_result;
    }

    //// begin protected interface

    // protected:

    function access_base(&$action, &$dbx, &$custom)
    {
        $this->m_action = &$action;
        $this->m_dbx = &$dbx;
        $this->m_custom = &$custom;

        $this->config();
    }

    function get_login_method()
    {
        $login_method = null;

        if (isset($this->m_config[MAP_AUTH_INTERFACE_DEF][$this->get_database_logical()]))
        {
            $login_method = LOGIN_METHOD_APP;
        }
        else if  (@dbx_base::get_engine() == ENGINE_OCI)
        {
            $login_method = LOGIN_METHOD_DIRECT;
        }
        else if ($this->m_action->get_login_method())
        {
            $login_method = $this->m_action->get_login_method();
        }

        return $login_method;
    }

    //// begin private interface

    // private:

    function config()
    {
        $access_inc = get_config_dir() . 'access.inc.php';

        include($access_inc);
    }

    function is_diagnostic_msg_enabled()
    {
        return $this->m_config[DIAGNOSTIC_MSG_ENABLE_FLAG] && $this->m_action->is_show_response();
    }

    function add_diagnostic_msg($msg)
    {
        if ($this->is_diagnostic_msg_enabled())
        {
            add_debug_msg($msg);
        }
    }

    function select_db_diagnostic($db_name, $select_result)
    {
        if ($this->is_diagnostic_msg_enabled())
        {
            $this->add_diagnostic_msg("select_db('$db_name') - attempt " . ($select_result === false ? 'failed' : 'SUCCEEDED'));
        }
    }

    function get_site_prefix($host = null)
    {
        if (!$host)
        {
            $host = strtolower($_SERVER['HTTP_HOST']);
        }

        return !empty($this->m_config[MAP_SITE_PREFIX][$host]) ? $this->m_config[MAP_SITE_PREFIX][$host] : '' ;
    }

    function get_database_logical($alternate_appid = null)
    {
        return $this->m_custom->get_database($alternate_appid);
    }

    function get_database_physical($alternate_appid = null, $host = null)
    {
        if (!$host)
        {
            $host = strtolower($_SERVER['HTTP_HOST']);
        }

        $logical = $this->get_database_logical($alternate_appid);
        $physical = $logical;

        if (!empty($logical))
        {
            $physical = !empty($this->m_config[MAP_DATABASE_PHYSICAL][$host][$logical]) ?
                        $this->m_config[MAP_DATABASE_PHYSICAL][$host][$logical] :
                        ($this->get_site_prefix() . $logical) ;
        }

        return $physical;
    }

    function get_username_physical($logical)
    {
        return $this->get_site_prefix() . $logical;
    }

    function connect_diagnostic(&$link, $hostname, $username, $auth_connect = false)
    {
        if ($this->is_diagnostic_msg_enabled())
        {
            $this->add_diagnostic_msg(($auth_connect ? 'auth_connect' : 'db_connect') . "(-h '$hostname', -u '$username') - attempt " . ($link === false ? 'failed' : 'SUCCEEDED'));
        }
    }

    function fetch_auth_interface_def()
    {
        assert('!isset($this->m_auth_interface_def)');
    }

    function auth_connect()
    {
        $host = strtolower($_SERVER["HTTP_HOST"]);

        $credentials = !empty($this->m_config[MAP_AUTH_CREDENTIALS][$host]) ?
                       $this->m_config[MAP_AUTH_CREDENTIALS][$host] :
                       array($this->m_config[ONE_FOR_ALL_HOSTNAME],
                             $this->get_username_physical($this->m_connect_as),
                             $this->m_config[ONE_FOR_ALL_PASSWORD]) ;

        $link = $this->m_dbx->connect($credentials[1], $credentials[2], $credentials[0]);

        $this->connect_diagnostic($link, $credentials[0], $credentials[1], true);

        return $link;
    }

    function validation_diagnostic($username, $user_validated)
    {
        if ($this->is_diagnostic_msg_enabled())
        {
            $msg = "user validation " . ($user_validated ? 'SUCCEEDED' : 'FAILED') .
                   " against " . $this->get_database_physical($this->m_auth_interface_def->database()) . '.' . $this->m_auth_interface_def->table() .
                   " for " . $this->m_auth_interface_def->field_username() . " = '$username'" .
                   " and " . $this->m_auth_interface_def->field_password() . " = '*******'";

            $this->add_diagnostic_msg($msg);
        }
    }

    function authenticate()
    {
        $u = $this->m_action->get_login_username();
        $p = $this->m_action->get_login_password();

        if (!empty($u) || !empty($p)) // detect if session login is occuring
        {
            $this->set_token(); // cancel any previous session login

            $link = $this->auth_connect();

            if ($link)
            {
                if ($this->m_auth_interface_def)
                {
                    $u = $this->m_dbx->escape_sql_term($u);
                    $p = $this->m_dbx->escape_sql_term($p);

                    $this->m_dbx->select_db($this->get_database_physical($this->m_auth_interface_def->database()));

                    $q = "select count(*) from " . $this->get_database_physical($this->m_auth_interface_def->database()) . '.' . $this->m_auth_interface_def->table() .
                         " where " . $this->m_auth_interface_def->field_username() . " = '$u'" .
                         " and " . $this->m_auth_interface_def->field_password() . " = " . $this->m_auth_interface_def->function_encrypt($p);

                    $v = $this->m_dbx->get_query_value($q);

                    $user_validated = 1 == $this->m_dbx->get_query_value($q);

                    $this->validation_diagnostic($u, $user_validated);

                    if ($user_validated)
                    {
                        $default_hash = $this->m_config[ONE_FOR_ALL_HASH];
                        $hash = $this->resolve($u, $default_hash);
                        $this->set_token($hash); // session login successful
                        $this->set_username($u);
                        $this->set_login_appid($this->m_action->get_appid());
                    }
                }

                $this->m_dbx->close();
            }
        }

        return $this->is_authenticated();
    }

    function set_token($token = null)
    {
        if (empty($token))
        {
            unset_session_var('token');
            $this->set_username();
            $this->set_login_appid();
        }
        else
        {
            set_session_var('token', $token);
        }
    }

    function get_token()
    {
        return get_session_var('token');
    }

    function set_username($username = null)
    {
        if (empty($username))
        {
            unset_session_var('username');
        }
        else
        {
            set_session_var('username', $username);
        }
    }

    function get_username()
    {
        return get_session_var('username');
    }

    function set_login_appid($login_appid = null)
    {
        if (empty($login_appid))
        {
            unset_session_var('login_appid');
        }
        else
        {
            set_session_var('login_appid', $login_appid);
        }
    }

    function get_login_appid()
    {
        return get_session_var('login_appid');
    }

    function resolve($map_key, $default_hash = null)
    {
        if (empty($this->m_config[MAP_HASH_CREDENTIALS][$this->m_config[ONE_FOR_ALL_HASH]]))
        {
            $this->m_config[MAP_HASH_CREDENTIALS][$this->m_config[ONE_FOR_ALL_HASH]] = array(null,
                                                                                             $this->m_config[ONE_FOR_ALL_HOSTNAME],
                                                                                             (!$this->get_site_prefix() || !$this->get_database_physical() ?
                                                                                              $this->get_username_physical(SYNAPP2) :
                                                                                              $this->get_database_physical()),
                                                                                             $this->m_config[ONE_FOR_ALL_PASSWORD]);
        }

        ////

        $result = null;

        if (!empty($map_key))
        {
            if (!$default_hash)
            {
                // map hash to connection credential
                //
                if (isset($this->m_config[MAP_HASH_CREDENTIALS][$map_key]))
                {
                    $result = $this->m_config[MAP_HASH_CREDENTIALS][$map_key];
                }
            }
            else
            {
                // map username to hash (of appropriate connection credential)
                //
                foreach ($this->m_config[MAP_HASH_CREDENTIALS] as $hash => $credentials)
                {
                    if ($credentials[0] === $map_key)
                    {
                        $result = $hash;
                        break;
                    }
                }

                if (!$result && isset($this->m_config[MAP_HASH_CREDENTIALS][$default_hash]))
                {
                    $result = $default_hash;
                }
            }
        }

        return $result;
    }

    function db_connect($force = false)
    {
        if ($force)
        {
            $this->m_link = false;
        }

        if ($this->m_link === false)
        {
            $credentials = $this->resolve($this->get_token());

            if (!empty($credentials))
            {
                $this->m_link = $this->m_dbx->connect($credentials[2], $credentials[3], $credentials[1]);
                $this->connect_diagnostic($this->m_link, $credentials[1], $credentials[2]);
            }
        }

        return $this->m_link !== false;
    }
};

class access_app extends access_base
{
    function fetch_auth_interface_def()
    {
        $this->m_auth_interface_def = isset($this->m_config[MAP_AUTH_INTERFACE_DEF][$this->get_database_logical()]) ?
                                      $this->m_config[MAP_AUTH_INTERFACE_DEF][$this->get_database_logical()] :
                                      new auth_interface_def($this->get_database_logical()) ;
    }

    function auth_connect()
    {
        $this->fetch_auth_interface_def();

        $this->m_connect_as = $this->get_site_prefix() ? $this->get_database_logical() : SYNAPP2 ;

        ////

        return parent::auth_connect();
    }
};

class access_synapp2 extends access_base
{
    function fetch_auth_interface_def()
    {
        $this->m_auth_interface_def = new auth_interface_def(SYNAPP2);
    }

    function auth_connect()
    {
        $this->fetch_auth_interface_def();

        $this->m_connect_as = SYNAPP2;

        ////

        return parent::auth_connect();
    }
};

class access_enterprise extends access_base
{
    function authenticate()
    {
        $u = $this->m_action->get_login_username();
        $p = $this->m_action->get_login_password();

        if (!empty($u) || !empty($p))
        {
            $this->set_token();

            $authenticated = false; // enterprise_authentication($u, $p);

            if ($authenticated)
            {
                $default_hash = $this->m_config[ONE_FOR_ALL_HASH];
                $hash = $this->resolve($u, $default_hash);
                $this->set_token($hash);
                $this->set_username($u);
                $this->set_login_appid(SYNAPP2);
            }
        }

        return $this->is_authenticated();
    }
};

class access_direct extends access_base
{
    function authenticate()
    {
        $u = $this->m_action->get_login_username();
        $p = $this->m_action->get_login_password();

        if (!empty($u) || !empty($p))
        {
            $this->set_token();

            $authenticated = true; // bypass authentication layer

            if ($authenticated)
            {
                $hash = array(null, $this->m_config[ONE_FOR_ALL_HOSTNAME], $u, $this->obfuscator($p));
                $this->set_token($hash);
                $this->set_username($u);
                $this->set_login_appid(SYNAPP2);
            }
        }

        return $this->is_authenticated();
    }

    function db_connect($force = false)
    {
        if ($force)
        {
            $this->m_link = false;
        }

        if ($this->m_link === false)
        {
            $credentials = $this->get_token();

            if (!empty($credentials))
            {
                $this->m_link = $this->m_dbx->connect($credentials[2], $this->obfuscator($credentials[3], true), $credentials[1]);

                $this->connect_diagnostic($this->m_link,  $credentials[1], $credentials[2]);

                if (!$this->m_link)
                {
                    $this->set_token();
                }
            }
        }

        return $this->m_link !== false;
    }

    function obfuscator($subject, $decode = false)
    {
        return $subject; // TODO: SECURITY: implement encode/decode
    }
};

////

class auth_interface_def
{
    // customize via instances of this class, not by changing the member initialization
    //
    var $m_database = '';
    var $m_table = 'users';
    var $m_field_username = 'username';
    var $m_field_password = 'password';
    var $m_function_encrypt = 'password'; // 'oldpassword'; '';

    function auth_interface_def($database,
                                $table = null,
                                $field_username = null,
                                $field_password = null,
                                $function_encrypt = null)
    {
        $this->database($database);
        $this->table($table);
        $this->field_username($field_username);
        $this->field_password($field_password);
        $this->function_encrypt('', $function_encrypt);
    }

    function database($database = null)
    {
        if (!empty($database))
        {
            $this->m_database = $database;
        }

        return $this->m_database;
    }

    function table($table = null)
    {
        if (!empty($table))
        {
            $this->m_table = $table;
        }

        return $this->m_table;
    }

    function field_username($field_username = null)
    {
        if (!empty($field_username))
        {
            $this->m_field_username = $field_username;
        }

        return $this->m_field_username;
    }

    function field_password($field_password = null)
    {
        if (!empty($field_password))
        {
            $this->m_field_password = $field_password;
        }

        return $this->m_field_password;
    }

    function function_encrypt($value, $function_encrypt = null)
    {
        if (!empty($function_encrypt))
        {
            $this->m_function_encrypt = $function_encrypt;
        }

        return $this->m_function_encrypt . "('$value')";
    }
};

////

define('LOGIN_METHOD_APP', 'app');
define('LOGIN_METHOD_SYNAPP2', SYNAPP2);
define('LOGIN_METHOD_DIRECT', 'direct');
define('LOGIN_METHOD_ENTERPRISE', 'enterprise');

define('DIAGNOSTIC_MSG_ENABLE_FLAG', '_diagnostic_msg_enable_flag_');

define('MAP_AUTH_CREDENTIALS', '_map_auth_credentials_');
define('MAP_HASH_CREDENTIALS', '_map_hash_credentials_');
define('MAP_AUTH_INTERFACE_DEF', '_map_auth_interface_def_');
define('MAP_DATABASE_PHYSICAL', '_map_database_physical_');
define('MAP_SITE_PREFIX', '_map_site_prefix_');

define('ONE_FOR_ALL', '_one_for_all_');

define('ONE_FOR_ALL_HOSTNAME', '_one_for_all_hostname_');
define('ONE_FOR_ALL_PASSWORD', '_one_for_all_password_');
define('ONE_FOR_ALL_HASH', '_one_for_all_hash_');

////

function &new_access(&$action, &$dbx, &$custom)
{
    $obj = null;

    $base = new access_base($action, $dbx, $custom);

    switch ($base->get_login_method())
    {
    case LOGIN_METHOD_APP:
        $obj = new access_app($action, $dbx, $custom);
        break;
    case LOGIN_METHOD_SYNAPP2:
    default:
        $obj = new access_synapp2($action, $dbx, $custom);
        break;
    case LOGIN_METHOD_DIRECT:
        $obj = new access_direct($action, $dbx, $custom);
        break;
    case LOGIN_METHOD_ENTERPRISE:
        $obj = new access_enterprise($action, $dbx, $custom);
        break;
    }

    $obj->authenticate();

    ////

    if ($action->get_appid() == SYNAPP2)
    {
        // TODO: parameterize qid/table and col_names according to access config
        $activity = array();
        $custom->set_order('users', 'username', $activity);
        $custom->set_validator('users', 'username', 'validate_username_synapp2', VALIDATOR_SECONDARY);
        $custom->set_validator('users', 'password', 'validate_password_synapp2');
    }

    ////

    return $obj;
}

////

function is_auth_appid($appid)
{
    global $g_access;

    ////

    $login_appid = $g_access->get_login_appid();

    return $login_appid == SYNAPP2 || $login_appid == $appid;
}

////

function validate_username_synapp2($col_name, $col_value, $display_name, &$col_vars)
{
    global $g_action;

    $msg = '';

    if (!empty($col_value))
    {
        if (preg_match('/[^a-z0-9_\.\-\@]+/i', $col_value) || $col_value != escape_sql_term($col_value))
        {
            $msg = "Username contains disallowed characters";
        }
        else
        {
            $update_filter = $g_action->is_update() ? (" and id != '" . $g_action->get_pk() . "'") : '' ;

            if (get_dbx()->get_query_value("select count(*) from users where username = '{$col_value}'{$update_filter}"))
            {
                $msg = "Username already exists";
            }
        }
    }

    return $msg;
}

function validate_password_synapp2($col_name, $col_value, $display_name, &$col_vars)
{
    global $g_action;

    $msg = '';

    if (!empty($col_value) && ($password_action = $g_action->m_process_args['password_action']))
    {
        if ($col_value != escape_sql_term($col_value))
        {
            $msg = "{$display_name} contains disallowed characters";
        }
        else
        {
            if (preg_match('/encrypt/i', $password_action))
            {
                $password_function = 'password'; // TODO: TBD: from access??

                $password_encrypted = get_dbx()->get_query_value("select {$password_function}('{$col_value}')");

                if (!empty($password_encrypted))
                {
                    $col_vars[$col_name]['v'] = $password_encrypted;
                }
                else
                {
                    $msg = "Unable to process {$display_name} value: {$password_function}('{$col_value}')";
                }
            }
        }
    }

    return $msg;
}

////

?>
