<?php

/* util.php - SynApp2 utility functions
**
   +----------------------------------------------------------------------+
   | SynApp2 Version 1                                                    |
   +----------------------------------------------------------------------+
   | Copyright (c) 2007 - 2010 Richard Howell. All rights reserved.       |
   +----------------------------------------------------------------------+
   | This source file is subject to version 1.01 of the SynApp2 license,  |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.synapp2.org/license/1_01.txt                              |
   | If you did not receive a copy of the SynApp2 license and are unable  |
   | to obtain it through the world-wide-web, please send a note to       |
   | license@synapp2.org so we can mail you a copy immediately.           |
   +----------------------------------------------------------------------+
   | Authors: Richard Howell                                              |
   |                                                                      |
   +----------------------------------------------------------------------+
**
** http://www.synapp2.org
**
** $Id: util.php,v 1.67 2010/10/23 18:47:13 richard Exp $
*/

function get_version($display = false, $force = false)
{
    global $g_test_mode;

    $version = '1.8.0';
    $release = 'beta_3'; // PHP version_compare() eval order: 'string not in list' < dev < alpha = a < beta = b < RC = rc < # < pl = p

    return $g_test_mode && !$force ? '0.0.0' : ($display && !empty($release) ? $version . ' (' . str_replace('_', ' ', $release) . ')' : $version . $release) ;
}

////

function session_begin()
{
    if (!session_id())
    {
        @session_start();
    }
}

session_begin();

function set_session_var($var_name, $var_value = null)
{
    if (!empty($var_name))
    {
        $_SESSION[$var_name] = $var_value;
    }
}

function get_session_var($var_name)
{
    return isset($_SESSION[$var_name]) ? $_SESSION[$var_name] : null ;
}

function unset_session_var($var_name)
{
    if (isset($_SESSION[$var_name]))
    {
        unset($_SESSION[$var_name]);
    }
}

////

$g_test_mode = false;

function set_test_mode($enable = true)
{
    global $g_test_mode;

    $g_test_mode = $enable ? true : false ;
}

////

$g_debug_msg = '';
$g_debug_msg_add_post_vars = true;

function add_debug_msg($mixed = null, $tag = null)
{
    global $g_debug_msg;
    global $g_debug_msg_add_post_vars;

    if ($g_debug_msg_add_post_vars)
    {
        $g_debug_msg_add_post_vars = false;
        fmt_debug($g_debug_msg, $_POST, '$_POST');
    }

    fmt_debug($g_debug_msg, $mixed, $tag);
}

function fmt_debug(&$markup, &$mixed, $tag)
{
    ob_start();

    if (!empty($tag))
    {
        print_r("\n");
        print_r($tag);
        print_r("\n");
    }

    if (!empty($mixed))
    {
        print_r($mixed);
    }

    $buf = encode_entities(ob_get_contents());

    ob_end_clean();

    if (!empty($buf))
    {
        $markup .= "<debug_msg>$buf</debug_msg>\n";
    }
}

function fmt_line(&$markup, $line)
{
    $markup .= $line . "\n";
}

////

function fmt_display_name($subject, $clip_id = true)
{
    $result = $subject;

    if (is_string($result))
    {
        $result = trim($result);

        if ($result == strtoupper($result))
        {
            $result = strtolower($result);
        }

        if ($clip_id)
        {
            $result = preg_replace('/(^id_|_id$)/i', '', $result);
        }

        $result = ucwords(str_replace('_', ' ', $result));
    }

    return $result;
}

////

function assert_handler($filename, $line, $expr)
{
    add_debug_msg("Expr: $expr\nFile: $filename\nLine: $line\n", '* Assertion Failed *');
}

assert_options(ASSERT_CALLBACK, 'assert_handler');
assert_options(ASSERT_ACTIVE, 1);
assert_options(ASSERT_WARNING, 0);
assert_options(ASSERT_QUIET_EVAL, 1);

////

require_once('dbx.php');

function &get_dbx()
{
    global $g_dbx;

    return $g_dbx;
}

function escape_sql_term($term)
{
    $dbx = get_dbx();

    return !empty($dbx) ? $dbx->escape_sql_term($term) : $term ;
}

////

function get_username()
{
    global $g_access;

    assert($g_access);

    return !empty($g_access) ? $g_access->get_username() : null ;
}

function get_database_physical($alternate_appid = null, $host = null)
{
    global $g_access;

    assert($g_access);

    return !empty($g_access) ? $g_access->get_database_physical($alternate_appid, $host) : '' ;
}

////

function get_array_element(&$array_var, $key_value)
{
    return isset($array_var[$key_value]) ? $array_var[$key_value] : null ;
}

function explode_csv($csv)
{
    return preg_split('/[\s]*[,][\s]*/', $csv);
}

////

$g_php_shutdown_action = null;

function set_php_shutdown_action($php_shutdown_action=null)
{
    global $g_php_shutdown_action;

    $g_php_shutdown_action = is_string($php_shutdown_action) && !empty($php_shutdown_action) ? $php_shutdown_action : null ;
}

function on_php_shutdown()
{
    global $g_php_shutdown_action;

    if (!empty($g_php_shutdown_action))
    {
        eval($g_php_shutdown_action);
    }
}

register_shutdown_function('on_php_shutdown');

////

$g_check_expr_msg = null;

function on_check_expr_fault($errno, $errstr)
{
    global $g_check_expr_msg;

    $g_check_expr_msg = $errstr;

    return true;
}

function on_check_expr_fatal()
{
    $feedback = 'No details are available';

    if (is_callable('error_get_last'))
    {
        $last_error = error_get_last(); // >= PHP5.2

        if (!empty($last_error['message']))
        {
            $feedback = $last_error['message'];
        }
    }

    $feedback = "<status name=\"message\">The expression triggers a fatal error: {$feedback}</status>";

    ////

    $response = '';

    fmt_line($response, '<' . '?xml version="1.0" encoding="utf-8" standalone="yes" ?' . '>');
    fmt_line($response, "<response>");
    fmt_line($response, "<version>" . get_version() . "</version>");
    fmt_line($response, "<authentication>succeeded</authentication>");
    fmt_line($response, "<authorization>succeeded</authorization>");
    fmt_line($response, "<adhoc>");
    fmt_line($response, '<feedback>');
    fmt_line($response, $feedback);
    fmt_line($response, '</feedback>');
    fmt_line($response, "</adhoc>");
    fmt_line($response, "</response>");

    ////

    header("Content-Type: text/xml");
    echo $response;
}

function check_expr($code_fragment, $is_raw = false)
{
    global $g_check_expr_msg;

    $g_check_expr_msg = null;

    ini_set('display_errors', '1'); // prevent PHP from registering HTTP 500 internal server error during eval()
    ini_set('html_errors', false);  // suppress markup

    $delim = $is_raw ? '' : '"' ;

    if (!@eval("return true || {$delim}" . $code_fragment . "{$delim};"))
    {
        $g_check_expr_msg = 'Unparsable';
    }
    else
    {
        set_error_handler('on_check_expr_fault');
        set_php_shutdown_action("on_check_expr_fatal();");

        @eval('$lvalue=' . $delim . $code_fragment . "{$delim};"); // fatal errors, such as undefined functions, cannot be trapped

        set_php_shutdown_action();
        restore_error_handler();
    }

    ini_restore('html_errors');
    ini_restore('display_errors');

    return $g_check_expr_msg; // null == ok
}

////

$g_entity_reference_map = null;

function encode_entities($subject)
{
    global $g_entity_reference_map;

    $result = null;

    if (isset($subject))
    {
        if (is_array($subject))
        {
            $result = array_map('encode_entities', $subject);
        }
        else if (htmlentities($subject, ENT_NOQUOTES) != $subject) // 10% faster to check all of $subject first
        {
            $result = '';

            $len = strlen($subject);

            for ($i = 0; $i < $len; $i++)
            {
                $c = $subject[$i];

                $encoded_entity = htmlentities($c, ENT_NOQUOTES);

                if ($encoded_entity && $encoded_entity != $c && ($n = ord($c)) > 127)
                {
                    $g_entity_reference_map[preg_replace(array('/^&/', '/;$/'), '', $encoded_entity)] = '&#' . $n . ';';
                }

                $result .= $encoded_entity;
            }
        }
    }

    return isset($result) ? $result : $subject ;
}

function decode_entities($subject)
{
    return utf8_decode($subject);
}

function get_entity_references()
{
    global $g_entity_reference_map;

    $result = '';

    if (!empty($g_entity_reference_map))
    {
        fmt_line($result, '<!DOCTYPE response [');

        foreach ($g_entity_reference_map as $entity_name => $entity_value)
        {
            fmt_line($result, "<!ENTITY {$entity_name} \"{$entity_value}\">");
        }

        fmt_line($result, ']>');
    }

    return $result;
}

////

function downloadFile($filename)
{
    if (!empty($filename) && file_exists($filename))
    {
        header("Pragma: public");
        header("Expires: 0");
        header("Cache-Control: must-revalidate, post-check=0, pre-check=0");
        header("Content-Type: application/force-download");
        header("Content-Type: application/octet-stream");
        header("Content-Type: application/download");
        header("Content-Disposition: attachment; filename=" . basename($filename) . ";");
        header("Content-Transfer-Encoding: binary");
        header("Content-Length: " . filesize($filename));

        readfile("$filename");
    }
}

function writeFileContents($output, $lines_out)
{
    // return file_put_contents($output, $lines_out); // php 5

    $result = false;

    $h = fopen($output, 'wb+');

    if ($h)
    {
        $wresult = true;

        if (!empty($lines_out))
        {
            foreach ($lines_out as $line)
            {
                if (fwrite($h, $line) === false)
                {
                    $wresult = false;
                    break; // error
                }
            }
        }

        $result = $wresult && fclose($h);
    }

    return $result;
}

////

function get_work_dir()
{
    return './';
}

function get_apps_dir()
{
    return get_work_dir() . '../';
}

function get_apps_relative_dir($relative_dir)
{
    return get_apps_dir() . "{$relative_dir}/";
}

function get_app_dir($appid)
{
    return $appid ? (get_apps_dir() . "{$appid}/") : get_work_dir() ;
}

function get_config_dir()
{
    return get_apps_dir() . '_config_/';
}

function get_temp_dir()
{
    return function_exists('sys_get_temp_dir') && is_dir(sys_get_temp_dir()) ? sys_get_temp_dir() : get_apps_dir() ;
}

////

function mkdir_app_dir($appid)
{
    $result = null;

    if (!empty($appid))
    {
        $apps_dir = get_apps_dir();

        if (is_dir($apps_dir))
        {
            $app_dir = get_app_dir($appid);

            if (!is_dir($app_dir))
            {
                mkdir($app_dir);
                chmod($app_dir, 0777);
            }

            if (is_dir($app_dir))
            {
                $result = $app_dir;
            }
        }
    }

    return $result;
}

////

?>
